"""Base Nosql abstract class container"""
from abc import ABC, abstractmethod

__author__ = 'ozgur'
__creation_date__ = '9.09.2019' '14:12'

from typing import Union, Dict, Any, List, Optional


class BaseNosqlConnector(ABC):
    """Base Nosql connector Class"""

    def __init__(self):
        self.tagprefix = "tag:"

    @abstractmethod
    def delete_table(self):
        """
        deletes table from db\n
        :return:
        """

    @abstractmethod
    def create_table(self):
        """
        creates a new table, delete first if exists \n
        :return:
        """

    @abstractmethod
    def dbsize(self) -> int:
        """
        returns the itemcount in db\n
        :return: item count
        """

    @abstractmethod
    def keys(self, pattern='*') -> List[str]:
        """
        returns keys list in db \n
        :param pattern: str urlpathregex
        :return: the keys in db
        """

    @abstractmethod
    def get(self, key: str) -> Optional[Dict[str, Any]]:
        """
        get value by key\n
        :param key: str
        :return: dict or None
        """

    @abstractmethod
    def get_multi(self, keylist: List[str]) -> Dict[str, Dict[str, Any]]:
        """
        get dictionaries by keylist\n
        :param keylist: list of strings
        :return: [objectdict]
        """

    @abstractmethod
    def get_all_as_list(self) -> List[Optional[Dict[str, Any]]]:
        """
        get all values in store \n
        :return: list
        """

    @abstractmethod
    def get_all_as_dict(self) -> Dict[str, Optional[Dict[str, Any]]]:
        """
        get all values in store \n
        :return: dict
        """

    @abstractmethod
    def upsert(self, key: str, value: Dict[str, Any]) -> bool:
        """
        get all values in store \n
        :return: list
        """

    @abstractmethod
    def upsert_multi(self, itemdict: Dict[str, Dict[str, Any]]):
        """
        upsert multiple items at once\n
        :param itemdict: dict of objects dictionaries
        :return: None
        """

    @abstractmethod
    def remove(self, key: str) -> bool:
        """
        removes the key from store\n
        :param key: str or list or tuple
        :return:
        """

    @abstractmethod
    def remove_keys(self, keys: List[str]) -> bool:
        """
        removes the key from store\n
        :param keys:
        :return:
        """

    @abstractmethod
    def flush(self):
        """
        Flushes current store\n
        :return: None
        """

    @abstractmethod
    def tags(self) -> List[str]:
        """
        returns tag list \n
        :return:
        """

    @abstractmethod
    def gettagkeys(self, tag: str) -> List[str]:
        """
        get all keys of given tag \n
        :param tag:
        :return:
        """

    @abstractmethod
    def gettag(self, tag: str) -> List[Optional[Dict[str, Any]]]:
        """
        get all dict items in given tag \n
        :param tag:
        :return: list of dictionaries
        """

    @abstractmethod
    def addtag(self, tag: str, key: str):
        """
        Adds tags to keys \n
        :param tag: list
        :param key: list
        :return: None
        """

    @abstractmethod
    def removefromtag(self, tag: str, key: Union[str, list]):
        """
        Remove items from lists \n
        :param tag: str
        :param key: str or list
        :return: None
        """



# -*- coding: utf-8 -*-
"""
Basic redis connector
"""
import json
from typing import Union, List, Dict, Any, Optional

import redis

from nosqlmodel.connectors.baseconnector import BaseNosqlConnector

__author__ = 'ozgur'
__creation_date__ = '9.09.2019' '10:07'

RHOST: str = "localhost"
RPORT: int = 6379


class RedisConnector(BaseNosqlConnector):
    """Redis connector for models. Can also be used as standalone lib too"""

    def __init__(self, dbnum: int = 0):
        self.redisdbnum = dbnum
        self.conn = redis.StrictRedis(host=RHOST, port=RPORT, decode_responses=True,
                                      db=self.redisdbnum)
        super().__init__()

    def delete_table(self):
        """
        deletes table from db\n
        :return:
        """
        self.flush()

    def create_table(self):
        """
        creates a new table, delete first if exists \n
        :return:
        """

    def dbsize(self) -> int:
        """
        returns the itemcount in db\n
        :return: item count
        """
        return len(self.keys())

    def keys(self, pattern='*') -> List[str]:
        """
        returns keys list in db \n
        :param pattern: str urlpathregex
        :return: the keys in db
        """
        return list(
            set(self.conn.keys(pattern=pattern)) - set(self.conn.keys(self.tagprefix + "*")))

    def get(self, key: str) -> Optional[Dict[str, Any]]:
        """
        get value by key\n
        :param key: str
        :return: dict or None
        """
        if not isinstance(key, str):
            raise TypeError("key must be str, got " + str(type(key)))
        retval = None
        resp = self.conn.get(key)

        if resp:
            resp = json.loads(resp)
            if not isinstance(resp, dict):
                retval = {key: resp}
            else:
                retval = resp

        return retval

    def get_multi(self, keylist: List[str]) -> Dict[str, Dict[str, Any]]:
        """
        get value by key\n
        :param keylist: list of strings
        :return: [objectdict]
        """
        retdict: Dict[str, Any] = {}
        retlist = self.conn.mget(keylist)

        for i, rstr in enumerate(retlist):
            # noinspection PyBroadException
            try:
                # retlist[i] = json.loads(rstr)
                retdict[keylist[i]] = json.loads(rstr)
            except Exception:  # pylint: disable=broad-except
                # retlist[i] = str(rstr)
                retdict[keylist[i]] = str(rstr)
        return retdict

    def get_all_as_list(self) -> List[Optional[Dict[str, Any]]]:
        """
        get all values in store \n
        :return: list
        """
        keys = self.keys()
        retlist = []
        for key in keys:
            retlist.append(self.get(key))

        return retlist

    def get_all_as_dict(self) -> Dict[str, Optional[Dict[str, Any]]]:
        """
        get all values in store \n
        :return: dict
        """
        keys = self.keys()
        retdct = {}
        for key in keys:
            retdct[key] = self.get(key)

        return retdct

    def upsert(self, key: str, value: Dict[str, Any]) -> bool:
        """
        get all values in store \n
        :return: list
        """
        key = str(key)
        if not isinstance(value, dict):
            raise TypeError("value must be dict but got" + str(type(value)))
        nvalue = json.dumps(value, ensure_ascii=False)
        return bool(self.conn.set(key, nvalue))

    def upsert_multi(self, itemdict: Dict[str, Dict[str, Any]]):
        """
        upsert multiple items at once\n
        :param itemdict: dict of objects dictionaries
        :return: bool
        """
        pipe = self.conn.pipeline()
        for key, item in itemdict.items():
            mvalue = item.copy()
            if "idkey" not in mvalue or not mvalue["idkey"]:
                mvalue["idkey"] = str(key)
            nvalue = json.dumps(mvalue, ensure_ascii=False)
            pipe.set(str(key), nvalue)
        pipe.execute()

    def remove(self, key: str) -> bool:
        """
        removes the key from store\n
        :param key: str or list or tuple
        :return:
        """
        key = str(key)
        ret = str(self.conn.delete(key))
        return {"1": True}.get(ret, False)

    def remove_keys(self, keys: List[str]) -> bool:
        """
        removes the key from store\n
        :param keys:
        :return:
        """
        for key in keys:
            self.conn.delete(key)
        return True

    def flush(self):
        """
        Flushes current store\n
        :return: None
        """
        self.conn.flushdb()

    def tags(self) -> List[str]:
        """
        returns tag list \n
        :return:
        """
        taglist = self.conn.keys(self.tagprefix + "*")
        retlist = []
        for tag in taglist:
            retlist.append(tag.replace(self.tagprefix, ""))
        return retlist

    def gettagkeys(self, tag: str) -> List[str]:
        """
        get all keys of given tag \n
        :param tag:
        :return:
        """
        memberids = self.conn.smembers(self.tagprefix + tag)
        return list(memberids)

    def gettag(self, tag: str) -> List[Optional[Dict[str, Any]]]:
        """
        get all dict items in given tag \n
        :param tag:
        :return: list of dictionaries
        """
        memberids = self.gettagkeys(tag)
        retlist = []
        for memberid in list(memberids):
            retlist.append(self.get(memberid))

        return retlist

    def addtag(self, tag: str, key: str):
        """
        Adds tags to keys \n
        :param tag: list
        :param key: list
        :return: None
        """
        if not (tag and key):
            raise TypeError
        self.conn.sadd(self.tagprefix + tag, key)

    def removefromtag(self, tag: str, key: Union[str, list]):
        """
        Remove items from lists \n
        :param tag: str
        :param key: str or list
        :return: None
        """
        tag = str(tag)
        # ! dont touch it. srem doesnt accept list !!!!!
        if isinstance(key, (list, tuple, set)):
            for k in key:
                self.conn.srem(self.tagprefix + tag, k)
        else:
            self.conn.srem(self.tagprefix + tag, key)



"""
Basic dynamodb connector
"""
import time
from decimal import Decimal
from typing import Union, Dict, Any, List, Optional

# noinspection Mypy
import boto3
# from boto3.resources.factory.dynamodb import Table
# noinspection Mypy
from boto3.dynamodb.conditions import Key  # noqa
# noinspection Mypy
from botocore.exceptions import ClientError  # noqa

from nosqlmodel.connectors.baseconnector import BaseNosqlConnector

__author__ = 'ozgur'
__creation_date__ = '9.09.2019' '10:07'

D_REGION_NAME = ""
D_AWS_ACCESS_KEY_ID = ""
D_AWS_SECRET_ACCESS_KEY = ""


class DynamoConnector(BaseNosqlConnector):
    """    Dynamodb connector    """

    def __init__(self, tablename: str):
        self.source = boto3.resource(
            'dynamodb',
            region_name=D_REGION_NAME,
            aws_access_key_id=D_AWS_ACCESS_KEY_ID,
            aws_secret_access_key=D_AWS_SECRET_ACCESS_KEY
        )
        self.tablename = tablename
        # self.table: Table = self.source.Table(tablename)  # pylint: disable=E1101
        self.table = self.source.Table(tablename)  # pylint: disable=E1101
        self.taglistkey = "tags"

        super().__init__()

    @staticmethod
    def _f_to_d(fval: float) -> Decimal:
        """
        float to decimal converter \n
        :param fval: float value
        :return: Decimal
        """
        return Decimal(str(fval))

    @staticmethod
    def _d_to_f(dval: Decimal) -> Union[float, int]:
        """
        Deciaml to float converter \n
        :param dval: Decimal value
        :return: float or int
        """
        try:
            if int(str(dval).split(".")[1]) == 0:
                retval = int(dval)
            else:
                # noinspection Mypy
                retval = float(dval)
        except IndexError:
            retval = int(dval)
        return retval

    def _d_to_f_dict(self, target: dict) -> dict:
        """
        checks items in dict and converts them to float if decimal \n
        :param target: raw dict
        :return: converted dict
        """
        if target:
            for key, value in target.items():
                if isinstance(value, Decimal):
                    target[key] = self._d_to_f(value)
        return target

    def delete_table(self):
        """
        deletes table from db\n
        :return:
        """
        try:
            self.table.delete()
            time.sleep(10)
        except ClientError:
            pass

    def create_table(self):
        """
        creates a new table, delete first if exists \n
        :return:
        """
        self.table = self.source.create_table(  # pylint: disable=E1101
            TableName=self.tablename,
            KeySchema=[
                {
                    'AttributeName': 'idkey',
                    'KeyType': 'HASH'  # Partition key
                },
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'idkey',
                    'AttributeType': 'S'
                },

            ],
            ProvisionedThroughput={
                'ReadCapacityUnits': 10,
                'WriteCapacityUnits': 10
            }
        )
        time.sleep(10)

    @staticmethod
    def issuccess(resp: dict) -> bool:
        """
        checks if request is success (returns 200)\n
        :param resp: boto response dict
        :return: bool
        """
        return resp.get("ResponseMetadata", {}).get("HTTPStatusCode", 0) == 200

    def keys(self, pattern='*') -> List[str]:
        """
        returns keys list in db \n
        :param pattern: str urlpathregex
        :return: the keys in db
        """
        retlist = []
        resp = self.table.scan(
            ProjectionExpression='idkey',
        )
        data = resp['Items']

        while 'LastEvaluatedKey' in resp:
            resp = self.table.scan(
                ExclusiveStartKey=resp['LastEvaluatedKey'],
                ProjectionExpression='idkey',
            )
            data.extend(resp['Items'])

        for rdict in data:
            val = str(rdict['idkey'])
            if not str(val).startswith(self.tagprefix):
                retlist.append(val)
        return retlist

    def dbsize(self) -> int:
        """
        returns the itemcount in db\n
        :return: item count
        """
        return len(self.keys())

    def get(self, key: str) -> Optional[Dict[str, Any]]:
        """
        get value by key\n
        :param key: str
        :return: dict or None
        """
        key = str(key)
        retval = self.table.get_item(
            Key={
                'idkey': key
            }
        ).get("Item", None)
        retval = self._d_to_f_dict(retval)

        return retval

    def get_multi(self, keylist: List[str]) -> Dict[str, Dict[str, Any]]:
        """
        get value by key\n
        :param keylist: list of strings
        :return: [objectdict]
        """
        keys = []
        for key in keylist:
            keys.append({"idkey": str(key)})

        request_object = {
            self.tablename: {
                'Keys': keys,
                'ConsistentRead': True
            }
        }

        response = self.source.batch_get_item(  # pylint: disable=no-member
            RequestItems=request_object,
            ReturnConsumedCapacity='TOTAL'
        )
        retlist = response['Responses'][self.tablename]
        retdict: Dict[str, Dict[str, Any]] = {}
        for itemdict in retlist:
            retdict[itemdict["idkey"]] = self._d_to_f_dict(itemdict)
        return retdict

    def get_all_as_list(self) -> List[Optional[Dict[str, Any]]]:
        """
        get all values in store \n
        :return: list
        """
        retlist: List[Optional[Dict[str, Any]]] = []
        for _, datadict in self.get_all_as_dict().items():
            retlist.append(datadict)

        return retlist

    def get_all_as_dict(self) -> Dict[str, Optional[Dict[str, Any]]]:
        """
        get all values in store \n
        :return: dict
        """

        resp = self.table.scan()
        data = resp['Items']
        retval: Dict[str, Optional[Dict[str, Any]]] = {}

        while 'LastEvaluatedKey' in resp:
            resp = self.table.scan(
                ExclusiveStartKey=resp['LastEvaluatedKey']
            )
            data.extend(resp['Items'])
        for ddict in data:
            if not str(ddict["idkey"]).startswith(self.tagprefix):
                for key, value in ddict.items():
                    if isinstance(value, Decimal):
                        ddict[key] = self._d_to_f(value)
                retval[str(ddict["idkey"])] = ddict

        return retval

    @classmethod
    def _get_upsert_check(cls, key: str, value: Dict[str, Any]) -> Dict[str, Any]:
        """
        checks item before insert and returns converted item\n
        :param key:
        :param value:
        :return:
        """

        if not isinstance(value, dict):
            raise TypeError("value must be dict but got" + str(type(value)))
        retval = value.copy()
        if "idkey" not in retval or not retval["idkey"]:
            retval["idkey"] = key

        for ikey, data in retval.items():
            if isinstance(data, float):
                retval[ikey] = cls._f_to_d(data)
        return retval

    def upsert(self, key: str, value: Dict[str, Any]) -> bool:
        """
        get all values in store \n
        :return: list
        """
        updict = self._get_upsert_check(key, value)
        resp = self.table.put_item(Item=updict)
        return self.issuccess(resp)

    def upsert_multi(self, itemdict: Dict[str, Dict[str, Any]]):
        """
        upsert multiple items at once\n
        :param itemdict: dict of objects dictionaries
        :return: bool
        """
        with self.table.batch_writer(overwrite_by_pkeys=['idkey']) as batch:
            for key, value in itemdict.items():
                updict = self._get_upsert_check(key, value)
                batch.put_item(updict)

    def remove(self, key: str) -> bool:
        """
        removes the key from store\n
        :param key: str or list or tuple
        :return:
        """
        resp = self.table.delete_item(
            Key={
                'idkey': str(key),
            }
        )
        return self.issuccess(resp)

    def remove_keys(self, keys: List[str]) -> bool:
        """
        removes the key from store\n
        :param keys:
        :return:
        """
        with self.table.batch_writer() as batch:
            for key in keys:
                batch.delete_item(Key={'idkey': str(key)})
        return True

    def flush(self):
        """
        Flushes current store\n
        :return: None
        """
        records = self.keys()
        records += [self.tagprefix + tag for tag in self.tags()]
        with self.table.batch_writer() as batch:
            for key in records:
                batch.delete_item(Key={'idkey': key})

    def tags(self) -> List[str]:
        """
        returns tag list \n
        :return:
        """
        fexppr = Key("idkey").begins_with(self.tagprefix)
        resp = self.table.scan(
            FilterExpression=fexppr,
            ProjectionExpression='idkey',
        )
        data = resp['Items']

        while 'LastEvaluatedKey' in resp:
            resp = self.table.scan(
                FilterExpression=fexppr,
                ProjectionExpression='idkey',
                ExclusiveStartKey=resp['LastEvaluatedKey']
            )
            data.extend(resp['Items'])
        for index, ddict in enumerate(data):
            data[index] = ddict

        return [rdict['idkey'].replace(self.tagprefix, "", 1) for rdict in data]

    def gettagkeys(self, tag: str) -> List[str]:
        """
        get all keys of given tag \n
        :param tag:
        :return:
        """
        tdict = self.get(self.tagprefix + tag)
        retval = tdict[self.taglistkey] if tdict else []
        return retval

    def gettag(self, tag: str) -> List[Optional[Dict[str, Any]]]:
        """
        get all dict items in given tag \n
        :param tag:
        :return: list of dictionaries
        """
        memberids = self.gettagkeys(tag)
        retlist = []
        for memberid in list(memberids):
            retlist.append(self.get(memberid))

        return retlist

    def addtag(self, tag: str, key: str):
        """
        Adds tags to keys \n
        :param tag: list
        :param key: list
        :return: None
        """
        if not (tag and key):
            raise TypeError
        tags = self.gettagkeys(tag)
        tags.append(key)
        tags = list(set(tags))
        ukey = self.tagprefix + tag
        self.upsert(ukey, {"idkey": ukey, self.taglistkey: tags})

    def removefromtag(self, tag: str, key: Union[str, list]):
        """
        Remove items from lists \n
        :param tag: str
        :param key: str or list
        :return: None
        """

        tags = list(set(self.gettagkeys(tag)))
        if isinstance(key, str):
            # noinspection PyBroadException
            try:
                tags.remove(key)
            except Exception:  # pylint: disable=broad-except
                pass
        else:
            for k in key:
                # noinspection PyBroadException
                try:
                    tags.remove(k)
                except Exception:  # pylint: disable=broad-except
                    pass

        ukey = self.tagprefix + tag
        self.upsert(ukey, {"idkey": ukey, self.taglistkey: tags})



"""Setup file for package"""
from setuptools import setup, find_packages

with open('README.md', encoding='utf8') as freadme:
    readme = freadme.read()

with open('LICENSE', encoding='utf8') as flicense:
    lcns = flicense.read()

setup(
    name='nosqlmodel',
    version='1.3.0',
    description='nosqlmodel is a NoSql ORM without relations. Easy way to create models with a '
                'nosql backend. Currently Redis and Dynamodb supported.',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/mozgurbayhan/nosqlmodel',
    author='Mehmet Ozgur Bayhan',
    author_email='mozgurbayhan@gmail.com',
    license="BSD",
    keywords='nosql orm redis dynamo',
    # py_modules=['nosqlmodel'],
    packages=find_packages(exclude=('tests', 'docs')),
    package_data={'': ['README.md']},
    include_package_data=True,
    python_requires='>=3.5',
    install_requires=[
        "simplejsonobject",
        "redis",
        "botocore",
        "boto3",
    ],
    project_urls={
        'Bug Reports': 'https://gitlab.com/mozgurbayhan/nosqlmodel/issues',
        'Funding': 'https://www.losev.org.tr/bagis/Bagis.html',
        'Say Thanks!': 'https://gitlab.com/mozgurbayhan/nosqlmodel',
        'Source': 'https://gitlab.com/mozgurbayhan/nosqlmodel'
    },

    classifiers=[
        'Topic :: Database :: Front-Ends',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
    ]

)
